# -*- coding: utf-8 -*-
"""
GUI functions for ukbc
"""
#colour scheme info
ukb_bg="#deeaec"
ukb_fg="#005f6f"
title_size=16
font="Arial"



def setup_loading_screen(app=None):
    app.addFileEntry("html_file")
    app.updateEntryDefault("html_file", "enter html file")
    app.addFileEntry("csv_file")
    app.updateEntryDefault("csv_file", "enter csv file")
    app.addStatusbar(fields=1)
    app.setStatusbar("No files entered", 0)
    


def setup_add_vars(app=None, ukbc=None):
    
    #show subwindow if already created, else build it
    try:
        app.showSubWindow("Adding Variables")
    except:
        #setting up subwindow
        app.startSubWindow("Adding Variables", modal=True)
        app.setStretch("both")
        app.setSticky("nesw")
        
        app.setBg(ukb_bg)
        app.setFont(title_size,font)
        app.addLabel("addVarTitle", "Adding Variables")
        app.setLabelBg("addVarTitle",ukb_bg)
        app.setLabelFg("addVarTitle", ukb_fg)
        
        #TODO: this information could be showed as a tabbed frame subdivided in to UKB showcase categories for ease of selection (see bottom of page 'TAB' for example code)
        # May need to parse the data showcase link into UKBC object...
        
        all_vars=ukbc.Vars[:]
        all_vars.sort()

        app.addListBox("var_options",all_vars)

        app.setListBoxMulti("var_options")
        app.setListBoxGroup("var_options")

         
        app.showSubWindow("Adding Variables")
    
    
#setting up select_multiple filtering option boxes 
def setup_select_multiple(app=None, ukbc=None, press=None):
    
    
    
    for opt in ["on_illness","off_illness"]:
        app.addLabel("Select "+opt,opt)
        i=ukbc.IllnessCodes['meaning'].tolist()
        i.sort()
        app.addListBox(opt,i)
        app.setListBoxMulti(opt)
        app.setListBoxGroup(opt)

        
    for opt in ["on_med","off_med"]:
        app.addLabel("Select "+opt,opt)
        m=ukbc.MedicationCodes['meaning'].tolist()
        m.sort()
        app.addListBox(opt, m)
        app.setListBoxMulti(opt)
        app.setListBoxGroup(opt)

                 
    #Add options to allocate excel output filename . . 
    app.addButtons(names=["Submit","Add Variables", "View Selection", "Cancel"], funcs=press)
        
        
def get_illness_med_selections(app=None):
        on_illness = app.getListBox("on_illness")
        off_illness = app.getListBox("off_illness")
        on_med = app.getListBox("on_med")
        off_med = app.getListBox("off_med")
        
        return on_illness, off_illness, on_med, off_med
    
    
def updateMeter(app=None, percentComplete=None):
    app.setMeter("progress", percentComplete)
    




# -*- coding: utf-8 -*-
"""
GUI setup for UKBCrunch

Requires 'appJar'
"""
import os
import UKBcrunch as UKBC 

from appJar import gui
from ukbc_functions import Filtering #Import filtering module:  Filtering requires the user to pass the ukbc object, followed by any desired options"""
from gui_functions.gui_functions import setup_loading_screen, setup_add_vars, setup_select_multiple
from gui_functions.gui_functions import get_illness_med_selections, updateMeter





#colour scheme info
ukb_bg="#deeaec"
ukb_fg="#005f6f"
title_size=16
font="Arial"
ukb_logo=os.path.dirname(__file__)+'/images/ukb_logo.gif'

def press(button):  

    if button == "Cancel":
        app.stop()
        
    elif button=="get_files":
        
        if len(app.getEntry("html_file"))>0:
            html_file=app.getEntry("html_file")
            csv_file=app.getEntry("csv_file")
            app.setStatusbar("Loading biobank data, please wait...", 0)
             
            app.infoBox("loading_ukbc",message="Loading UKBC object, will take ~3minutes")
            
            global ukbc #defining ukbc as global object
            ukbc=UKBC.UKBcrunch(html_file=html_file,csv_file=csv_file)
                
            app.infoBox("loaded_ukbc",message="Succesfully loaded UKBC object!")
            
            app.clearStatusbar(field=0)
            
            #This should remove file entry but doesn't - Q submitted to appJAr github
            app.removeFileEntry("html_file")
            app.removeFileEntry("csv_file")

            app.removeButton("get_files")
     
            #currently only provides functionality for 'select_multiple; filter
            setup_select_multiple(app=app,ukbc=ukbc,press=press)

             
    #Implemeting add vars option
    elif button=="Add Variables":        
        setup_add_vars(app=app, ukbc=ukbc)
    
      
    #display selections    
    elif button=="View Selection":
        
        #getting illness/med selections ..
        on_illness, off_illness, on_med, off_med = get_illness_med_selections(app)
        med_illness_msg=("Here are your current selections: \n\n ON ILLNESS: {0} \n OFF ILLNESS: {1} \n ON MED: {2} \n OFF MED {3} \n").format(on_illness,off_illness,on_med,off_med)
        
        #getting var selections..
        try:
            var_selections=app.getListBox("var_options")
        except:
            message=med_illness_msg+"No additional variables selected. Go back and choose 'Add Variables'"
        else:
            message=med_illness_msg+("Variables: {0}").format(var_selections)
            
        
        app.infoBox("selections",message)
        
    #generating df
    elif button=="Submit":
        
        #define output location
        fname=app.saveBox(title="Save file as ...", fileExt=None) 
        
        if len(fname)<1:
            fname='no_filename_given'
        
        #getting illness/med selections ..
        on_illness, off_illness, on_med, off_med = get_illness_med_selections(app)

        #getting var selections..
        try:
            var_selections=app.getListBox("var_options")
        except:
            var_selections=None
       
        
        #Creating excel file
        app.infoBox("loading_df","Your excel file is being created. You will be notified when it's ready (~3mins)")
        df=Filtering.select_by_multiple(ukbc, on_med=on_med, off_med=off_med, on_illness=on_illness, off_illness=off_illness)#creating df

        #if no illness or meds chosen then generate a full dataframe
        if all(len(v)<1 for v in [on_illness, off_illness, on_med, off_med]):
            df=ukbc.extract_many_vars()


        if var_selections is not None and df is not None:
            df=Filtering.add_vars(ukbc, in_df=df, add_vars=var_selections)#adding vars
        
        if df is not None:            
        #saving list of bulk fields requested.
            if len(ukbc.bulk_request)>0:
                outtxt = open(fname+'_bulk_fields.txt', 'w')
                for subj in df['eid']:
                    for item in ukbc.bulk_request:
                        outtxt.write("%s %s\n" % (subj, item))
                outtxt.close()
            
            df_decoded=Filtering.code2word(ukbc, df=df)
            df.to_excel(fname+'.xls')
            df_decoded.to_excel(fname+'_decoded.xls')
            app.infoBox("df","Your output was saved to: "+fname+".xls")
            
        else:
            app.infoBox("no_df","Your query returned no results")
            



"""      Setting up gui interface         """
# create a GUI variable called app
app = gui("UKBcrunch")
app.setLogLevel("CRITICAL")#prevents warning outputs
#app.setIcon(ukb_logo) # doesnt work atm
app.addImage("LOGO",ukb_logo)
app.setBg(ukb_bg)
app.setFont(title_size,font)



"""If UKBC object doesnt exist then must be loaded """
try:
    ukbc   
except:
    setup_loading_screen(app=app)
    app.addButton("get_files", press)


else:    
    app.addLabel("ukbc_found", "ukbc object detected, no need to reload!",colspan=2)
    #currently only provides functionality for 'select_multiple; filter
    setup_select_multiple(app=app, ukbc=ukbc, press=press)

     


# start the GUI
def start():       
    app.go()





# -*- coding: utf-8 -*-
"""


UKBcrunch , a simple tool for crunching the UKBioBank data.

Created Sept 2017 
@author: Joe Necus (NCL, UK)
@contact: j.necus2@ncl.ac.uk



"""

import pandas as pd
import bs4 #beautifulsoup4 package
import re  
import urllib3
import urllib2
import numpy as np
import os
from datetime import datetime


class UKBcrunch():
    """ A class to crunch UK Biobank data.
    
    To initialise:
        
        UKB_data=UKBcrunch(html_filepath, csv_filepath) 
  

    class variables
        html_file = path to html
        soup = parsed tables from html
        Vars = all variables extracted in data frame
        Eids = all Eids (encoded ID) extracted in data frame
    
        
    """
    
    special_char = "~`!@#$%^&*()_-+={}[]:>;',</?*-+" # needed to handle special chars in variable names
    sub_link  = 'http://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id='

    
    def __init__(self, html_file=None, csv_file=None):
                
        # Get filepath        
        if html_file is not None and csv_file is not None:
            #Construct the path to the html file
            self.html_file = html_file
            self.csv_file = csv_file
        else:
            ValueError("Must give location to folder containing html & csv")
        
        
        self.coding3_file=os.path.dirname(__file__)+'/coding_files/coding3.tsv'
        self.coding4_file=os.path.dirname(__file__)+'/coding_files/coding4.tsv'
        self.coding6_file=os.path.dirname(__file__)+'/coding_files/coding6.tsv'
        
        # Number of cases in csv
        self.N = len(pd.read_csv(self.csv_file,low_memory=False))
        
        #stores bulk variable code requests
        self.bulk_request=[]

        # Parse html
        # This enables easy searching of the html file for information on variables
        self.soup = self.makeSoup()
        
        #Time/date variables (initialised to arbitrary dates)
        self.date_format = "%Y-%m-%d"
        self.end_follow_up = "2016-02-15"
        self.start_follow_up = "2006-05-10"
        self.Time_end = datetime.strptime(self.end_follow_up, self.date_format)
        
        #Gathering all Variables from html file 
        self.Vars = self.gather_all_vars()['names']
        self.Data_types = self.gather_all_vars()['types']

        #Illness codes
        self.IllnessCodes=self.get_illness_codes()
        self.CancerCodes=self.get_cancer_codes()
        
        #Medication codes
        self.MedicationCodes=self.get_medication_codes() 


        #Gathering all eids from html file 
        self.Eids = self.gather_all_eids()
        self.OK = False
        if self.Eids is not None:
            self.OK = True
        if not self.OK:
            print 'error - failed to get Eids'
            self.OK = False
            return
        
        #Gathering all assesment centre visit dates ([53.0] 1st visit [53-1] 2nd visit [53-2] Imaging visit)
        
	if 'Date of attending assessment centre' in self.Vars:
            self.assess_dates = self.get_assessment_dates()

        
        
        
    """ Methods defined here onwards """
        

    def makeSoup(self):
        """Parse the html into a nested data structure"""
        f = open(self.html_file, 'r').read()
        soup = bs4.BeautifulSoup(f, 'html.parser')
        return soup
        

    def gather_all_vars(self):
        """Read all variable names in the table and return"""
        
        allrows = self.soup.findAll('tr')
        res = []
        data_type = []
        for t in allrows:
            re_string = 'nowrap;\">(.*?)</span></td><td rowspan=\"(\d+)\">(.*?)</td></tr>'
            res1=re.search(re_string,str(t))
            if not res1 is None:
                res1 = res1.group(0)
                ## get variable data type
                x1,y1,z1=res1.partition('nowrap;\">')
                xx1,yy1,zz1=z1.partition('</span>')
                data_type.append(xx1)
                ## get variable name
                x,y,z = zz1.partition('">')
                xx,yy,zz = z.partition('</td></tr>')
                if xx.find('<br>') > -1:
                    t = xx.find('<br>')
                    xx = xx[0:t]
                res.append(xx)
        res2 = []
        for x in res:
            if x.find('<br/>') >-1:
                 t = x.find('<br/>')
                 xx = x[0:t]
                 res2.append(xx)
            else:
                 res2.append(x)
        res = res2
        return {'names':res, 'types':data_type}

    def gather_all_eids(self):
        """Return all the EIDs"""
        # data frame of EIDs
        filename = self.csv_file 
        if filename == None:
            return None
        EIDs = pd.read_csv(filename, usecols=['eid'], nrows=self.N)
        return EIDs
    
    def get_assessment_dates(self):
        # data frame of EIDs
        var = 'Date of attending assessment centre'
        Ds = self.extract_variable(var)
        return Ds   


    #TODO: Extract variable currently accesses UKB showcase. May wish to edit this to access the html file (or some other more stable resource..)
    def extract_variable(self, variable=None):
        """Extract a single set of variable IDs given a single variable name as a list or string"""
        
        ### extract fields from supplied .html file
        allrows = self.soup.findAll('tr') # all rows in html file
        
        
        ## search variable string for problematic characters
        symbols = UKBcrunch.special_char

        varlist = list(variable)
        lvarlist = len(varlist)
        newvar = []
        for v in range(0, lvarlist):
            if varlist[v] in symbols:
                newvar.extend(['\\', varlist[v]])
            else:
                newvar.extend([varlist[v]])
        variable = ''.join(newvar)
                
        userrows = [t for t in allrows if re.search('>'+variable+'<',str(t))]
        
        if not userrows:
            userrows = [t for t in allrows if re.search('>'+variable+'.<',str(t))]
            
        #TODO:This fails if an unrecognised variable is found. This function should throw an error if so to enable easy detection of error
        
        userrows_str = str(userrows[0])

        IDs = [] # extract raw ids (e.g. 31 (Sex)) IDs related to variables
        
        # extract all variable names 
        match1 = re.search('id=(\d+)',userrows_str)
        if match1:
            IDs.append(match1.group(1))
            
        var_names = variable

        ## Retrieve all associated columns with variables names 
        Sub_list = {}
        
        
        for idx in IDs:

            key = []
                       
            #accessing row in html source file according to reference showcase link
            for link in self.soup.find_all('a', href=UKBcrunch.sub_link+idx):               
                tmp = str(link.contents[0].encode('utf-8'))#return associated code (e.g. sex (31-0))
                key.append(tmp)
            
            
                         
                ### Below I am trialling accessing the data showcase to determine whether the current field is 'Bulk' data or not        
                #Delete if necessary #                
                #opening data showcase link and parsing
                try:
                    page = urllib2.urlopen(UKBcrunch.sub_link+idx).read()
                except:
                    pass #page not acccesible
                else:
                    soup = bs4.BeautifulSoup(page, 'html.parser')            
                
                    for link in soup.find_all('td', class_="txt_blu"):#locating part of html where bulk field is located
                        bulk_type=str(link.contents[0].encode('utf-8'))
                    
                        #If current variable is deemed to be 'bulk', then save this into list of bulk vars
                        if bulk_type=='Bulk':
                            
                            bulk_code=tmp.replace('-','_').replace('.','_')#generating bulk code (contains underscores instead of . or -)
                            self.bulk_request.append(bulk_code)
                ### Trialling above ###
            
            
            
            Sub_list[var_names]=key
            
            
        my_range = Sub_list[var_names]
        my_range.append('eid') # Encoded anonymised participant ID
        filename = self.csv_file
        everything = pd.read_csv(filename,usecols=my_range,nrows=self.N)
        
            
        return everything
        
    
    def extract_many_vars(self, keywords=None, spaces=False, visit=None,instance=None, demographics=True):
        """Extracts variables for several pre-specified var. names 
        returns one single df with eids and each variables as columns"""
        
        if type(keywords)==str: # if string given (single var) then convert to list
            ls=[]
            ls.append(keywords)
            keywords=ls[:]
        
        #Returning standard demographic information
        if demographics==True:
            if keywords==None:
                keywords=[]
                
            if 'Sex' in self.Vars and 'Age when attended assessment centre' in self.Vars:
                        keywords.extend(['Sex','Age when attended assessment centre'])
            
            
            # including other demographics can cause trouble later down the line. E.g. if visit1 is selected then it will make it difficult to obtain age at visit 2, it will be hard to.
            #'Body mass index (BMI)','Age when attended assessment centre','Ethnic background','Sex'])
        
        #Extract vars
        main_Df = pd.DataFrame(columns = ['eid'])
        main_Df['eid'] = self.Eids['eid']
        for var in keywords:
            DBP = self.extract_variable(variable=var)
            if spaces:
                b,k,a = var.partition(' ')
                var = b
            DBP = self.rename_columns(DBP, var)           
            
            main_Df = pd.merge(main_Df,DBP,on='eid',how='outer')
        
        if visit is not None:
                main_Df=self.select_visit(df=main_Df,visit=visit, demographics=demographics)
                
        if instance is not None:
                main_Df=self.select_instance(df=main_Df,instance=instance)
        
        
        return main_Df
     
    
    def select_visit(self, df=None,visit=None,demographics=False):
        """returns dataframe containing only selected visit"""
        ls=['eid']#include eid column
        
        #extract columns with '_(visit)'
        for var in df.columns:
            res = re.search('(.*?)_'+str(visit)+'.(\d+)',var)
            if not res is None:
                ls.append(res.group(0))
        
        if visit > 0 and demographics is True:            
            ls.append('Sex_0.0')#appending sex, which is only included as visit 0       
        
        return df.loc[:,ls]#return df with list of selected columns
    
    
    def select_instance(self, df=None,instance=None):
        # returns dataframe containing only selected instnace e.g. '0.(instance)'
        ls=['eid']#include eid column
        
        #extract columns with '_(visit)'
        for var in df.columns:
            res = re.search('(.*?)_(\d+).'+str(instance)+'(.*)',var)
            if not res is None:
                ls.append(res.group(0))
               
        return df.loc[:,ls]#return df with list of selected columns
       
    


    def show_related_vars(self, keyword=None):
        
        if type(keyword)!=list: #if not list then make list
            ls=[]
            ls.append(keyword)            
            keyword=ls[:]
        
        keywords='|'.join('({0})'.format(i) for i in keyword)

        matches = [t for t in self.Vars if re.search(keywords,t,re.IGNORECASE)]
        
        if len(matches)<1:
            print('No matches found')
        
        return matches


    def get_demographics(self, visit=None):
        """ Returns simple dataframe containing demographics information """
        return self.extract_many_vars()
    
    def gather_related_vars(self, keyword=None,visit=None,instance=None,demographics=True):
    # Extracts all variables 'related' to a keyword variable or list of keyword variables. Note: this function will ignore case.
    # Returns single df 

           
        if type(keyword)!=list: #if not list then make list
            ls=[]
            ls.append(keyword)            
            keyword=ls[:]
            
        keywords='|'.join('({0})'.format(i) for i in keyword)
        
        matches = [t for t in self.Vars if re.search(keywords,t,re.IGNORECASE)]#searching all vars for those containing keyword
        

        if len(matches)<1:
            print('No matches found')
            data=None
            
        #may need to include if len ==1, self.extract_singl_var ..
        else:
            data=self.extract_many_vars(keywords=matches,visit=visit,instance=instance,demographics=demographics)
        
        return data
        
    
    def vars_by_visits(self, col_names=None,visit=0):
        # returns variables names associated with initial assessment (0), 1st (1) and 2nd (2) re-visit
        # 1st (1) and 2nd (2) re-visit
        V1 =[]
        for var in col_names:
            res = re.search('(.*?)-'+str(visit)+'.(\d+)',var)
            if not res is None:
                V1.append(res.group(0))
        return V1 
    
    
    def get_cancer_codes(self):
        """Returns data coding convention from coding3 files (cancer encoding)"""
        
        Cancer_codes=pd.read_table(self.coding3_file,index_col='coding',usecols=['coding','meaning'])
        Cancer_codes.drop(-1,inplace=True)#drop '-1' codings
        return Cancer_codes
        
    
    def get_illness_codes(self):
        """Returns data coding convention from coding6 files (illness encoding)"""
        
        noCancer_codes=pd.read_table(self.coding6_file,index_col='coding',usecols=['coding','meaning'])
        noCancer_codes.drop(-1,inplace=True)#drop '-1' codings
        return noCancer_codes
    
    def get_medication_codes(self):
        """Returns data coding convention from coding4 file"""
        med_codes=pd.read_table(self.coding4_file,index_col='coding')
        return med_codes
    
    
    def get_illness_summary(self, in_df=None, Cancer=False):    
        """Returns information regarding the frequency of various illnesses in the current data
        sample (if in_df is not provided then the whole  'ukbc' object dataset will be summarised"""
        
        if Cancer==True:
            k='Cancer code, self-reported'#setting column keywords
            illness_codes=self.CancerCodes
        else:
            k='Non-cancer illness code, self-reported'
            illness_codes=self.IllnessCodes
        
        #1. Get dataframe with all medication information
        if in_df is not None:
            df=in_df.copy()
        else:
            df=self.gather_related_vars(keyword=k)            
        
        illness_columns=df.filter(regex=k).columns
        
        out_df=pd.DataFrame()
        
        #2. For each illness in illness codes count the frequencies (separately for each sex)
        for sex,sex_df in df.groupby('Sex_0.0'):
            for code in illness_codes.index:
                    bool_df=sex_df.loc[:,illness_columns]==int(code)
                    freq=(bool_df.sum(axis=1)>0).sum()
                    out_df.loc[illness_codes.loc[code,'meaning'],'IllnessFrequencySex'+str(sex)]=freq       
        
   
        return out_df 
    
    
    def get_medication_summary(self, in_df=None):    
        """Returns information regarding the frequency of various medications in the current data
        sample (if in_df is not provided then the whole  'ukbc' object dataset will be summarised"""
        
        k='Treatment/medication code'#setting column keyword
        medication_codes=self.MedicationCodes 
        
        #1. Get dataframe with all medication information
        if in_df is not None:
            df=in_df.copy()
        else:
            df=self.gather_related_vars(keyword=k)            
        
        
        medication_columns=df.filter(regex=k).columns
       
        out_df=pd.DataFrame()
        
        #2. For each illness in illness codes count the frequencies (separately for each sex)
        for sex,sex_df in df.groupby('Sex_0.0'):
            for code in medication_codes.index:
                    bool_df=sex_df[medication_columns]==int(code)
                    freq=(bool_df.sum(axis=1)>0).sum()
                    out_df.loc[medication_codes.loc[code,'meaning'],'MedicationFrequencySex'+str(sex)]=freq
        return out_df 
    

    
    def rename_columns(self, df=None,key=None,option_str=True):
        # rename the columns of a data frame with something sensible
        col_names = df.columns.tolist()
        col_new = ['eid']
        for k in range(3):
            V0 = self.vars_by_visits(col_names,k)
            for v in V0:
                b,k,a = v.partition('-')
                if option_str:
                    col_new.append(key+'_'+a)
                else:
                    col_new.append(key)
        df_new = pd.DataFrame(columns=col_new)
        for c in range(len(col_new)):
            df_new[col_new[c]] = df[col_names[c]]
        return df_new 
    
    
    
    
    



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  6 15:18:24 2017

@author: Joe Necus
"""

from setuptools import setup,find_packages

setup(name='ukbc',
      version='1.1',
      description='Interface to faciliate filtering of UKBioBank data',
      include_package_data=True,
      url='http://github.com/jnecus/ukbcrunch',
      author='Joe Necus',
      author_email='j.necus2@ncl.ac.uk',
      license='MIT',
      install_requires=['appjar'],
      zip_safe=False,
      packages=find_packages(),
      package_data={'ukbc.coding_files': ['*'],
                        'ukbc.images': ['*'],
				'ukbc.data': ['*']})



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Oct  5 12:20:49 2017

@author: b5013167
"""


""" Filtering: Functions which enable filtering of UK Biobank data.

    Currently includes (Oct 17):
        
        -Filter_by_multiple
        -code2word
        -get_matched_sample (WIP)
    


"""

import pandas as pd
import numpy as np
import time
import sys
import re

def cleanup(out_df,subset=None):
    """ Drop all nan rows, columns and duplicates"""
     
    if out_df is None:
        print("Dataframe is empty")
    elif len(out_df)<1:
        out_df=None
        print("Dataframe is empty")
    else:    
        #out_df.dropna(axis=0, how='all',inplace=True, subset=subset)# drop if all columns in row are nan
        out_df.dropna(axis=1, how='all',inplace=True)# drop if all rows in  column are nan
        out_df.drop_duplicates(inplace=True)#drop duplicate rows
    
    return out_df



def convert_list(i):
    """convert strings to lists"""
    ls=[]
    if type(i)==str:
        ls.append(i)
        o=ls[:]
    else:
        o=i[:]
    
    return o
    

def code2word(ukbc, df=None):    
    """  Converts numerical values in dataframe to text based upon UK biobank coding """
    
    if df is not None: 
        
        #dataframe to be returned
        out_df=df.copy()
       
        #Converting medication coding     
        med_cols=out_df.filter(regex='Treatment/medication code').columns.tolist()#coding columns..
        
        #If medication code columns exist, then continue
        if len(med_cols)>0:
            codes_dict=ukbc.MedicationCodes.to_dict()['meaning']
            for col in med_cols:
                out_df[col].replace(codes_dict, inplace=True)
                
        
        #Converting illness coding 
        illness_cols=[]
        
        illness_cols.extend(out_df.filter(regex='Non-cancer illness code, self-reported').columns.tolist())
    
        
        #If illness code columns exist, then continue
        if len(illness_cols)>0:
            codes_dict=ukbc.IllnessCodes.to_dict()['meaning']
            codes_dict = {int(k):v for k,v in codes_dict.items()}#converting keys to int
            for col in illness_cols:
                out_df[col].replace(codes_dict, inplace=True)
               
        cancer_cols=[]
        cancer_cols.extend(out_df.filter(regex='Cancer code, self-reported').columns.tolist())
        
        #If cancer code columns exist, then continue
        if len(cancer_cols)>0:
            codes_dict=ukbc.CancerCodes.to_dict()['meaning']
            codes_dict = {int(k):v for k,v in codes_dict.items()}#converting keys to int
            for col in cancer_cols:
                out_df[col].replace(codes_dict, inplace=True)
                
                
        #TODO: continue to implement other types of coding (see UK biobank showcase)...    
        
        #e.g.1 ethnic background.
        #e.g. 2	Job SOC coding (data coding 2) (req download of coding file and assignment to ukbc object...)
    return out_df



def select_by_illness(ukbc, in_df=None, illness=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with given string of list of illnesses
         NOTE: If a list [A,B,C] is supplied, then an instance is treated as if A OR B OR C is present """

    
    #If string given then convert to list
    illness=convert_list(illness)

    
    ill_codes=[]
    cancer_codes=[]
    for i in illness:        
        #Gather illness code(s) codes must be list of int
        x=ukbc.IllnessCodes[ukbc.IllnessCodes==i]#illness codes
        x.dropna(inplace=True)   
        
        y=ukbc.CancerCodes[ukbc.CancerCodes==i]#cancer codes
        y.dropna(inplace=True)
        
        if len(x)<0 and len(y)<0:        
            raise Exception('Error: cannot find the illness: {0}. Please ensure that you are using medications exactly as specified in ''IllnessCodes or CancerCodes'''.format(i))
        
        #TODO: Currently this function will treat illnessess & cancer which share the same code as the same
        #How to overcome, include 'Cancer' flag' in function
        elif len(x)>0:
            ill_codes.append(int(x.index.tolist()[0]))
        else:
            cancer_codes.append(int(y.index.tolist()[0]))            
        
       
    illness_df=ukbc.extract_many_vars(keywords=['Cancer code, self-reported','Non-cancer illness code, self-reported'],visit=visit,instance=instance)
    
    #selecting only illness columns
    cancer_cols_of_interest=illness_df.filter(regex='Cancer code, self-reported').columns.tolist()
    illness_cols_of_interest=illness_df.filter(regex='Non-cancer illness code, self-reported').columns.tolist()
    
    
    #If the code is contained within df then append to out_df...
    include_rows=[]
    
    for c in ill_codes:        
            #Is code in dataframe?
            bool_df=illness_df[illness_cols_of_interest]==c        
            #If any element in a row is true, then include this row          
            inc=bool_df[bool_df.any(axis=1)].index.tolist()  
            include_rows.extend(inc)
            
    for c in cancer_codes:        
            #Is code in dataframe?
            bool_df=illness_df[cancer_cols_of_interest]==c        
            #If any element in a row is true, then include this row          
            inc=bool_df[bool_df.any(axis=1)].index.tolist()  
            include_rows.extend(inc)
            
    
    #Append included rows
    out_df=illness_df.loc[include_rows,:]   
    
    cols_of_interest2=[]
    cols_of_interest2.extend(out_df.filter(regex='Cancer code, self-reported').columns.tolist())
    cols_of_interest2.extend(out_df.filter(regex='Non-cancer illness code, self-reported').columns.tolist())
    
    #merge incoming dataframe
    if in_df is not None:
        cols_to_use=in_df.columns.difference(out_df.columns).tolist()
        cols_to_use.append('eid')
        out_df = out_df.merge(in_df[cols_to_use],on='eid',how='inner')
    
    out_df=cleanup(out_df,subset=cols_of_interest2)    
    
    
    return out_df


def select_by_illness_free(ukbc, in_df=None, illness=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with no record of a given list of medications
    NOTE: If a list [A,B,C] is supplied, then an instance is treated as if A OR B OR C is present """
    
    #If string given then convert to list
    illness=convert_list(illness)
    
    
    ill_codes=[]
    cancer_codes=[]
    for i in illness:        
        #Gather illness code(s) codes must be list of int
        x=ukbc.IllnessCodes[ukbc.IllnessCodes==i]#illness codes
        x.dropna(inplace=True)   
        
        y=ukbc.CancerCodes[ukbc.CancerCodes==i]#cancer codes
        y.dropna(inplace=True)
        
        if len(x)<0 and len(y)<0:        
            raise Exception('Error: cannot find the illness: {0}. Please ensure that you are using medications exactly as specified in ''IllnessCodes or CancerCodes'''.format(i))
            
        elif len(x)>0:
            ill_codes.append(int(x.index.tolist()[0]))
        else:
            cancer_codes.append(int(y.index.tolist()[0]))     
    
    
    illness_df=ukbc.extract_many_vars(keywords=['Cancer code, self-reported','Non-cancer illness code, self-reported'],visit=visit,instance=instance)
    
    #selecting only illness columns
    cols_of_interest=[]
    cols_of_interest.extend(illness_df.filter(regex='Cancer code, self-reported').columns.tolist())
    cols_of_interest.extend(illness_df.filter(regex='Non-cancer illness code, self-reported').columns.tolist())
    
    cancer_cols_of_interest=illness_df.filter(regex='Cancer code, self-reported').columns.tolist()
    illness_cols_of_interest=illness_df.filter(regex='Non-cancer illness code, self-reported').columns.tolist()
    
    
    #If the code is contained within df then append to out_df...
    dont_include_rows=[]
    
    for c in ill_codes:        
            #Is code in dataframe?
            bool_df=illness_df[illness_cols_of_interest]==c        
            #If any element in a row is true, then include this row          
            dont_inc=bool_df[bool_df.any(axis=1)].index.tolist()  
            dont_include_rows.extend(dont_inc)
            
    for c in cancer_codes:        
            #Is code in dataframe?
            bool_df=illness_df[cancer_cols_of_interest]==c        
            #If any element in a row is true, then include this row          
            dont_inc=bool_df[bool_df.any(axis=1)].index.tolist()  
            dont_include_rows.extend(dont_inc)
            
                
    #Append included rows
    out_df=illness_df.drop(dont_include_rows)
    
    
    cols_of_interest2=[]
    cols_of_interest2.extend(out_df.filter(regex='Cancer code, self-reported').columns.tolist())
    cols_of_interest2.extend(out_df.filter(regex='Non-cancer illness code, self-reported').columns.tolist())
        
    #merge incoming dataframe
    if in_df is not None:
        
        #finding column which are present in in_df but not out_df, merging with out_df
        cols_to_use=in_df.columns.difference(out_df.columns).tolist()
        cols_to_use.append('eid')
        out_df = out_df.merge(in_df[cols_to_use],on='eid',how='inner')
        
    
    out_df=cleanup(out_df,subset=cols_of_interest2)
    
    return out_df


def select_by_medication(ukbc, in_df=None, medication=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with given string of list of medications """
    
    #If string given then convert to list
    medication=convert_list(medication)
    
    #Gather medication code(s) codes must be list of int
    codes=[]
    for i in medication:        
        x=ukbc.MedicationCodes[ukbc.MedicationCodes==i]
        x.dropna(inplace=True)
        if len(x)<1:
            raise Exception('Error: cannot find the medication {0}. Please ensure that you are using medications exactly as specified in ''MedicationCodes'''.format(i))
        else:
            codes.append(int(x.index.tolist()[0]))
    
    
    medication_df=ukbc.extract_many_vars(keywords=['Treatment/medication code'],visit=visit,instance=instance)
    
    #selecting only illness columns -ugly workaround (basically we want to avoid including the demographics columns here (which are returned by the 
    #above function by default. If an add_demographics() function is implemented then demographics=False can be implemeted above to circumvent this))
    cols_of_interest=[]
    cols_of_interest.extend(medication_df.filter(regex='Treatment/medication code').columns.tolist())
    
    #If the code is contained within df then append to out_df...
    include_rows=[]
    
    for c in codes:        
        #Is code in dataframe?
        bool_df=medication_df[cols_of_interest]==c        
        #If any element in a row is true, then include this row          
        inc=bool_df[bool_df.any(axis=1)].index.tolist()  
        include_rows.extend(inc)
    
    #Append included rows
    out_df=medication_df.loc[include_rows,:]
            
    cols_of_interest2=out_df.filter(regex='Treatment/medication code').columns.tolist()     
    
    #merge incoming dataframe
    if in_df is not None:
        cols_to_use=in_df.columns.difference(out_df.columns).tolist()
        cols_to_use.append('eid')
        out_df = out_df.merge(in_df[cols_to_use],on='eid',how='inner')
    
    
    out_df=cleanup(out_df,subset=cols_of_interest2)

    return out_df



def select_by_medication_free(ukbc, in_df=None, medication=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with no record of a given list of medications 
         NOTE: If a list [A,B,C] is supplied, then an instance is treated as if A OR B OR C is present """
    
    #If string given then convert to list
    medication=convert_list(medication)

    
    
    codes=[]
    for i in medication:        
        #Gather medication code(s) codes must be list of int
        x=ukbc.MedicationCodes[ukbc.MedicationCodes==i]
        x.dropna(inplace=True)
        if len(x)<1:
            raise Exception('Error: cannot find that medication: {0}. Please ensure that you are using medications exactly as specified in ''MedicationCodes'''.format(i))
        else:
            codes.append(int(x.index.tolist()[0]))
    
    
    medication_df=ukbc.extract_many_vars(keywords=['Treatment/medication code'],visit=visit,instance=instance)
    
    #selecting only illness columns -ugly workaround (basically we want to avoid including the demographics columns here (which are returned by the 
    #above function by default. If an add_demographics() function is implemented then demographics=False can be implemeted above to circumvent this))
    cols_of_interest=[]
    cols_of_interest.extend(medication_df.filter(regex='Treatment/medication code').columns.tolist())
    
    
    dont_include_rows=[]
    for c in codes:        
        #Is code in dataframe?
        bool_df=medication_df[cols_of_interest]==c
        
        #If any element in a row is true, then DONT include this row
        dont_inc=bool_df[bool_df.any(axis=1)].index.tolist()  
        
        dont_include_rows.extend(dont_inc)
    
    #Append included rows
    out_df=medication_df.drop(dont_include_rows)
    
    
    cols_of_interest2=out_df.filter(regex='Treatment/medication code').columns.tolist()  
        
    #merge incoming dataframe
    if in_df is not None:
        
        #finding column which are present in in_df but not out_df, merging with out_df
        cols_to_use=in_df.columns.difference(out_df.columns).tolist()
        cols_to_use.append('eid')
        out_df = out_df.merge(in_df[cols_to_use],on='eid',how='inner')
        
    
    out_df=cleanup(out_df,subset=cols_of_interest2)
    
    return out_df








def select_by_medications(ukbc, in_df=None, on_medication=None, off_medication=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with given string of list of medications 
        NOTE: If a list [A,B,C] is supplied, then an instance is treated as if A OR B OR C is present """
    
    if on_medication is not None and off_medication is not None:        
        on_df=select_by_medication(ukbc=ukbc,in_df=in_df, medication=on_medication, visit=visit, instance=instance)
        off_df=select_by_medication_free(ukbc=ukbc,in_df=in_df, medication=off_medication, visit=visit, instance=instance)
        
        #Find common 'eids' between on and off dfs and merge
        #finding column which are present in in_df but not out_df, merging with out_df
        cols_to_use=on_df.columns.difference(off_df.columns).tolist()
        cols_to_use.append('eid')
        out_df = off_df.merge(on_df[cols_to_use],on='eid',how='inner')
    
    elif on_medication is not None and off_medication is None:
        out_df=select_by_medication(ukbc=ukbc,in_df=in_df, medication=on_medication, visit=visit, instance=instance)

    elif on_medication is None and off_medication is not None:
        out_df=select_by_medication_free(ukbc=ukbc,in_df=in_df, medication=off_medication, visit=visit, instance=instance)
    
    else:        
        ValueError('Error: you must specify either on_medication OR off_medication')
        out_df=None


    return out_df

def select_by_illnesses(ukbc, in_df=None, on_illness=None, off_illness=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with given string of list of illnesses
    NOTE: If a list [A,B,C] is supplied, then an instance is treated as if A OR B OR C is present """
    
    if on_illness is not None and off_illness is not None:        
        on_df=select_by_illness(ukbc=ukbc,in_df=in_df, illness=on_illness, visit=visit, instance=instance)
        off_df=select_by_illness_free(ukbc=ukbc,in_df=in_df, illness=off_illness, visit=visit, instance=instance)
        
        #Find common 'eids' between on and off dfs and merge. Finding column which are present in in_df but not out_df, merging with out_df
        cols_to_use=on_df.columns.difference(off_df.columns).tolist()
        cols_to_use.append('eid')
        out_df = off_df.merge(on_df[cols_to_use],on='eid',how='inner')
    
    elif on_illness is not None and off_illness is None:
        out_df=select_by_illness(ukbc=ukbc,in_df=in_df, illness=on_illness, visit=visit, instance=instance)

    elif on_illness is None and off_illness is not None:
        out_df=select_by_illness_free(ukbc=ukbc,in_df=in_df, illness=off_illness, visit=visit, instance=instance)
    
    else:        
        ValueError('Error: you must specify either on_illness OR off_illness')
        out_df=None


    return out_df



#TODO insert 'in_df' functionality into here
def select_by_multiple(ukbc, on_med=None,off_med=None, on_illness=None, off_illness=None, visit=None, instance=None):    
    """  Returns a dataframe containing subjects with given illnesses and medications.
         NOTE: If a list [A,B,C] is supplied, then an instance is treated as if A OR B OR C is present """
    
    
    to_merge=[]#collecting df's to merge

    
    if on_med is not None and len(on_med)>0:
        on_med_df=select_by_medication(ukbc=ukbc, medication=on_med, visit=visit, instance=instance)
        to_merge.append(on_med_df)
        
    if off_med is not None and len(off_med)>0:
        off_med_df=select_by_medication_free(ukbc=ukbc,medication=off_med, visit=visit, instance=instance)
        to_merge.append(off_med_df)
        
    if on_illness is not None and len(on_illness)>0:
        on_ill_df=select_by_illness(ukbc=ukbc, illness=on_illness, visit=visit, instance=instance)
        to_merge.append(on_ill_df)
        
    if off_illness is not None and len(off_illness)>0:
        off_ill_df=select_by_illness_free(ukbc=ukbc,illness=off_illness, visit=visit, instance=instance)
        to_merge.append(off_ill_df)
    
    
    #if any dataframes in 'to_merge' list are empty then drop them
    temp=to_merge[:]
    to_merge=[]
    for i in temp:
        if i is not None:
            if len(i)>0:
                to_merge.append(i)
    

    #Merging dataframes in to_merge list. Must be a simpler, cleaner way. Currenlty deciding which columns to use is v.messy . . .
    
    #if only one input option
    if len(to_merge)==1:
        out_df=to_merge[0].copy()
    
    #if two options selected
    elif len(to_merge)==2:
        cols1=to_merge[1].columns.difference(to_merge[0].columns).tolist()
        cols1.append('eid')
        out_df=to_merge[0].merge(to_merge[1][cols1],on='eid',how='inner')
    
    elif len(to_merge)==3:
        cols1=to_merge[1].columns.difference(to_merge[0].columns).tolist()
        cols1.append('eid')
        cols2=to_merge[2].columns.difference(to_merge[1].columns).tolist()
        cols2.append('eid')
        
        out_df=to_merge[0].merge(to_merge[1][cols1],on='eid',how='inner').merge(to_merge[2][cols2],on='eid',how='inner')
    
    elif len(to_merge)==4:
        cols1=to_merge[1].columns.difference(to_merge[0].columns).tolist()
        cols1.append('eid')
        cols2=to_merge[2].columns.difference(to_merge[1].columns).tolist()
        cols2.append('eid')
        cols3=to_merge[3].columns.difference(to_merge[2].columns).tolist()
        cols3.append('eid')
        
        out_df=to_merge[0].merge(to_merge[1][cols1],on='eid',how='inner').merge(to_merge[2][cols2],on='eid',how='inner').merge(to_merge[3][cols3],on='eid',how='inner')
        
    else:
        out_df=None
        #no illness/medication options chosen or results found 

    out_df=cleanup(out_df)
    
    return out_df

 
def add_vars(ukbc, in_df=None, add_vars=None, visit=None, instance=None):
    """ Adds selected variables to the current dataframe """
    
    if in_df is None or add_vars is None:
        ValueError('Error: you have failed to provide in_df and/or add_vars')
        out_df=None
        
    #extract variables #deleted: (but we dont want demographics columns)
    variables_df=ukbc.extract_many_vars(keywords=add_vars, visit=visit, instance=instance, demographics=False)
    
    #extract rows based on eids from in_df and all columns (except eid)
    in_eids=in_df['eid'] # gather encoded subject IDs
    rows=variables_df['eid'].isin(in_eids)
    variables_df=variables_df.loc[rows,:]
    
    #select only columns unique to variables_df
    cols_to_use=variables_df.columns.difference(in_df.columns).tolist()
    cols_to_use.append('eid')
    
    #join dataframes (variables_df and in_df)
    out_df=in_df.merge(variables_df[cols_to_use],on='eid',how='inner')
         
    out_df=cleanup(out_df)
        
    return out_df




def get_illness_code(ukbc, text_in):
    """Returns illness code
     given text input"""
     
    codes_dict=ukbc.IllnessCodes
    code=codes_dict[codes_dict['meaning']==text_in].index.tolist()[0]

    return code

#TODO: add functions which add variables to a new 'addvariables' document
#Work in progress . . .
def initial_diagnosis(ukbc, in_df=None, illness=None, visit=None, instance=None):
    """ Adds a column labelling Year and Age of first diagnosis for selected illnesses.
        Illness is expected as text, e.g. 'depression', (not corresponding code)"""
    
    
    out_df=in_df.copy()
    
    #getting initial diagnosis year and age of participans
    for v in ['Year','Age of participant']:
    
        #If string given then convert to list
        illness=convert_list(illness)
    
        """find code (or string) in in_df"""
        
        #get non cancer illness columns
        cols=out_df.filter(regex='Non-cancer illness code, self-reported')
        cols=cols.columns.tolist()
        
        
        #1. check that interpolated diagnosis year column exists, if not then request it
        if len(out_df.filter(regex='Interpolated '+v+' when non-cancer illness first diagnosed').columns)<1:
            out_df=add_vars(ukbc, in_df=out_df, add_vars='Interpolated '+v+' when non-cancer illness first diagnosed')
        
    
    
        #for each illness identify column
        for i in illness:
            
                    
            #get code for illness
            illness_code=get_illness_code(ukbc,i)
                   
            #creating dict to store year of diagnosis variable
            var_name=i+'_'+v+'_first_diagnosed'
            res_dict={'eid':[],var_name:[]}
            
            #iterrate over each row in df
            for row in out_df.iterrows():
                            
                #get columns with matching code
                matching_columns=row[1][row[1]==illness_code].index.tolist()
                
                
                #TODO: #NOTE only one value for 'INterpolated year..' wil be generated. TODO: If two different values exist then the mean will be taken
                if len(matching_columns)>0:
                    
                    c=matching_columns[0]
                       
                    found=re.findall('(.*)_(\d+).(\d+)',c)
                    visit=found[0][1]
                    instance=found[0][2]
                    
                    #extract intial diagnosis year based upon year and instance
                    initial_diagnosis_year=row[1]['Interpolated '+v+' when non-cancer illness first diagnosed_'+str(visit)+'.'+str(instance)]
                    eid=row[1]['eid']
                    
                    #create variable
                    res_dict['eid'].append(eid)
                    res_dict[var_name].append(initial_diagnosis_year) 
                     
                        
            new_df=pd.DataFrame.from_dict(res_dict)
            
            #for each illness merge latest new_df with in_df on 'eid' 
            out_df=out_df.merge(new_df,on='eid',how='outer')
        
    out_df=cleanup(out_df)
        
    return out_df





#Work in progress . . .
def get_matched_sample(ukbc, to_match=None, off_illness=None, off_med=None, visit=None, num_samples=1):    
    """ This function aims to return a dataframe returning a matched sample according with
        same M/F and same age distribution and allows the user to specificy illnesses NOT
        to be contained within the sample
        
        The returned sample will differ each time as the sampling df is shuffled"""
    
    if visit is None:
        sys.exit('You must provide a visit so that the sample can be age matched for the target visit')
        

    
    
    #getting dataframe free from illness (to gather matched sample from)
    if off_illness is not None and off_med is not None:
        off_ill_df=select_by_illness_free(ukbc,illness=off_illness,visit=visit)
        off_med_df=select_by_medication_free(ukbc, medication=off_med, visit=visit)
        df=off_ill_df.merge(off_med_df, how='inner')
    elif off_illness is not None:
        df=select_by_illness_free(ukbc,illness=off_illness,visit=visit)
    elif off_med is not None:
        df=select_by_medication_free(ukbc, medication=off_med,visit=visit)
    else:
        df=ukbc.gather_related_vars()
        
    #for each sample  
    matched_samples=[]
    for sample in range(num_samples):
        
        #shuffling healthy df
        df=df.sample(frac=1)
        
        #df to store out matched sample
        out_df=pd.DataFrame()
        
        #Gathering M/F age information for current distribution
        currentSexAge=[]
        for row in to_match.iterrows():
            currentSexAge.append((row[1]["Age when attended assessment centre_"+str(visit)+".0"],row[1]["Sex_0.0"]))
            
        #could remove timeout
        timeout = time.time() + 60 #times out after 1 minute
    
    
        #gathering matched sample from 'df'
        covered_rows=[]
        
        #cycling through each requirement, e.g.30yrs Female...
        for targetSexAge in currentSexAge:
            
            moveOn=False
            
            
            if time.time()>timeout:
                sys.exit('Error: Couldn''t find a sample matching your AGE criteria')
        
    
            #checking each row in the ill/med free df
            for i,row in enumerate(df.iterrows()):
            
                
                #while a match has not been found
                if moveOn is False:
                
                    #if current row has not already been selected
                    if i not in covered_rows:
                        
                        #if current row matches requirements                    
                        if (row[1]["Age when attended assessment centre_"+str(visit)+".0"] == targetSexAge[0]) & (row[1]["Sex_0.0"] == targetSexAge[1]):
                            
                            #append row to out df
                            out_df=out_df.append(row[1])
                            
                            #make sure this row isn't re-used
                            covered_rows.append(i)
                
                            #Pass onto next target sexAge
                            moveOn=True

        matched_samples.append(out_df)
    
    
    return matched_samples